/*
*  TINET (TCP/IP Protocol Stack)
*
*  Copyright (C) 2001-2012 by Dep. of Computer Science and Engineering
*                   Tomakomai National College of Technology, JAPAN
*
*  L쌠҂́Cȉ (1)`(4) ̏CFree Software Foundation
*  ɂČ\Ă GNU General Public License  Version 2 ɋL
*  qĂ𖞂ꍇɌC{\tgEFAi{\tgEFA
*  ς̂܂ށDȉjgpEEρEĔzziȉC
*  pƌĂԁj邱Ƃ𖳏ŋD
*  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
*      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
*      XR[hɊ܂܂Ă邱ƁD
*  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
*      pł`ōĔzzꍇɂ́CĔzzɔhLgip
*      ҃}jAȂǁjɁCL̒쌠\C̗pщL
*      ̖ۏ؋Kfڂ邱ƁD
*  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
*      płȂ`ōĔzzꍇɂ́C̏𖞂ƁD
*    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
*        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
*  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
*      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
*
*  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
*  TOPPERSvWFNǵC{\tgEFAɊւāC̓Kp\
*  ܂߂āCȂۏ؂sȂD܂C{\tgEFA̗pɂ蒼
*  ړI܂͊ԐړIɐȂ鑹QɊւĂC̐ӔC𕉂ȂD
*
*  @(#) $Id$
*/

#include <string.h>

#ifdef TARGET_KERNEL_ASP

#include <kernel.h>
#include <sil.h>
#include <t_syslog.h>
#include "kernel_cfg.h"
#include "tinet_cfg.h"

#endif	/* of #ifdef TARGET_KERNEL_ASP */

#ifdef TARGET_KERNEL_JSP

#include <s_services.h>
#include <t_services.h>
#include "tinet_id.h"

#endif	/* of #ifdef TARGET_KERNEL_JSP */

#include <tinet_defs.h>
#include <tinet_config.h>

#include <net/if.h>
#include <net/if_loop.h>
#include <net/if_ppp.h>
#include <net/ethernet.h>
#include <net/ppp_ipcp.h>
#include <net/net.h>
#include <net/net_endian.h>
#include <net/net_buf.h>
#include <net/net_timer.h>
#include <net/net_count.h>

#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcp_var.h>
#include <netinet/udp_var.h>
#include <netinet/ip_igmp.h>

#include <net/if_var.h>
#include <net/net_var.h>

#ifdef SUPPORT_IGMP

static T_IGMP_ENTRY igmp_groups[NUM_IGMP_ENTRY];
static T_IN4_ADDR igmp_send_addrs[NUM_IGMP_ENTRY];

static bool_t igmp_send_report_v2(T_IN4_ADDR dst);
static uint16_t igmp_get_timer(T_IGMP_ENTRY *entry);

ER igmp_set_loop(T_UDP_CEP *cep, uint8_t optval)
{
	switch (optval) {
	case 0:
		cep->igmp_loopback = false;
		return E_OK;
	case 1:
		cep->igmp_loopback = true;
		return E_OK;
	}

	return E_PAR;
}

ER igmp_get_loop(T_UDP_CEP *cep, uint8_t *optval)
{
	if (optval == NULL)
		return E_PAR;

	*optval = cep->igmp_loopback ? 1 : 0;

	return E_OK;
}

ER igmp_set_ttl(T_UDP_CEP *cep, uint8_t optval)
{
	cep->igmp_ttl = optval;

	return E_OK;
}

ER igmp_get_ttl(T_UDP_CEP *cep, uint8_t *optval)
{
	if (optval == NULL)
		return E_PAR;

	*optval = cep->igmp_ttl;

	return E_OK;
}

ER igmp_set_if(T_UDP_CEP *cep, T_IN4_ADDR *optval)
{
	if (optval == NULL)
		return E_PAR;

	cep->igmp_mcaddr = *optval;

	return E_OK;
}

ER igmp_get_if(T_UDP_CEP *cep, T_IN4_ADDR *optval)
{
	if (optval == NULL)
		return E_PAR;

	*optval = cep->igmp_mcaddr;

	return E_OK;
}

ER igmp_add_membership(T_UDP_CEP *cep, T_IP_MREQ *optval)
{
	T_IGMP_ENTRY *entry = NULL;
	int		i;
	uint16_t	min;
	T_IFNET	*ifp = IF_GET_IFNET();
	ER ret = E_NOMEM;

	if ((optval->imr_interface != IPV4_ADDRANY)
		&& (optval->imr_interface != ifp->in4_ifaddr.addr))
		return E_PAR;

	syscall(wai_sem(SEM_IGMP_GROUP_LOCK));

	for (i = 0; i < NUM_IGMP_ENTRY; i++) {
		entry = &igmp_groups[i];
		if (entry->timer == -1) {
			entry->timer = igmp_get_timer(entry);
			entry->ip_addr = optval->imr_multiaddr;
			ret = E_OK;
			break;
		}
	}

	syscall(sig_sem(SEM_IGMP_GROUP_LOCK));

	return ret;
}

ER igmp_drop_membership(T_UDP_CEP *cep, T_IP_MREQ *optval)
{
	T_IGMP_ENTRY *entry = NULL;
	int		i;
	uint16_t	min;
	ER ret = E_PAR;
	T_IFNET	*ifp = IF_GET_IFNET();

	if ((optval->imr_interface != IPV4_ADDRANY)
		&& (optval->imr_interface != ifp->in4_ifaddr.addr))
		return E_PAR;

	syscall(wai_sem(SEM_IGMP_GROUP_LOCK));

	for (i = 0; i < NUM_IGMP_ENTRY; i++) {
		entry = &igmp_groups[i];
		if ((entry->timer >= 0) && (entry->ip_addr == optval->imr_multiaddr)) {
			entry->timer = -1;
			entry->ip_addr = 0;
			ret = E_OK;
			break;
		}
	}

	syscall(sig_sem(SEM_IGMP_GROUP_LOCK));

	return ret;
}

ER igmp_set_ip_msfilter(T_UDP_CEP *cep, const T_IP_MSFILTER *optval)
{
	return E_PAR;
}

ER igmp_get_ip_msfilter(T_UDP_CEP *cep, T_IP_MSFILTER *optval)
{
	return E_PAR;
}

uint_t igmp_input(T_NET_BUF **inputp, uint_t *offp, uint_t *nextp)
{
	T_NET_BUF	*input = *inputp;
	T_IGMPV2_HDR	*igmph;
	T_IN4_ADDR	addr;
	uint_t		len, align;
	T_IP4_HDR	*ip4h;
	int i;
	T_IGMP_ENTRY *entry;
	bool_t ret;

	NET_COUNT_IGMP(net_count_igmp.in_octets,
		input->len - GET_IF_IP4_HDR_SIZE(input));
	NET_COUNT_IGMP(net_count_igmp.in_packets, 1);

	/* ICMP wb_̒`FbNB*/
	if (input->len < IF_IP4_IGMP_HDR_SIZE) {
		NET_COUNT_IGMP(net_count_igmp.in_err_packets, 1);
		NET_COUNT_MIB(igmp_stats.igmpInErrors, 1);
		goto buf_rel;
	}

	igmph = (T_IGMPV2_HDR *)(input->buf + *offp);

	/* 4 INebgẼf[^ */
	len = input->len - *offp;
	align = (len + 3) >> 2 << 2;

	/* 4 INebgE܂ŃpfBOŖ߂B*/
	if (align > len)
		memset((uint8_t*)input->buf + input->len, 0, (size_t)(align - len));

	/* `FbNTvZB*/
	if (in_cksum(igmph, align) != 0) {
		NET_COUNT_IGMP(net_count_igmp.in_err_packets, 1);
		goto buf_rel;
	}

	/* bZ[W̌^ɂ蕪򂷂B*/
	switch (igmph->type) {
	case IGMP_MEMBERSHIP_QUERY:
		ip4h = GET_IP4_HDR(input);
		ret = false;

		/* Q̃O[vAhX̃NG[̏ꍇ */
		syscall(wai_sem(SEM_IGMP_GROUP_LOCK));

		for (i = 0; i < NUM_IGMP_ENTRY; i++) {
			entry = &igmp_groups[i];
			if ((entry->timer >= 0) && (entry->ip_addr == addr)) {
				entry->resptime = igmph->time;
				ret = true;
				break;
			}
		}

		syscall(sig_sem(SEM_IGMP_GROUP_LOCK));

		if (ret) {
			/* o[Vbv|[g𑗐M */
			igmp_send_report_v2(ip4h->src);
		}
		break;
	}

buf_rel:
	syscall(rel_net_buf(input));
	return IPPROTO_DONE;
}

static bool_t igmp_send_report_v2(T_IN4_ADDR dst)
{
	T_IGMPV2_HDR	*igmph;
	T_NET_BUF		*output;
	uint_t			len, align;

	if (in4_get_datagram(&output, (uint_t)(IF_IP4_IGMP_HDR_SIZE + len), 0,
		&dst, NULL, IPPROTO_IGMP, IP4_DEFTTL,
		NBA_SEARCH_ASCENT, TMO_IGMP_OUTPUT) != E_OK)
		return false;

	/* IGMP wb_ݒ肷B*/
	igmph = GET_IGMPV2_HDR(output, IF_IP4_IGMP_HDR_OFFSET);
	igmph->type = IGMP_MEMBERSHIP_REPORT_V2;
	igmph->time = 0;
	igmph->addr = dst;

	/* 4 INebgẼf[^ */
	align = (len + 3) >> 2 << 2;

	/* 4 INebgE܂ŃpfBOŖ߂B*/
	if (align > len)
		memset((uint8_t*)GET_IGMP_SDU(output, IF_IP4_IGMP_HDR_OFFSET) + IP4_HDR_SIZE + len,
			0, (size_t)(align - len));

	/* `FbNTvZB*/
	igmph->sum = 0;
	igmph->sum = in_cksum(igmph, (uint_t)(IF_IP4_IGMP_HDR_SIZE + align));

	/* MB*/
	NET_COUNT_IGMP(net_count_igmp.out_octets,
		output->len - GET_IF_IP4_HDR_SIZE(output));
	NET_COUNT_IGMP(net_count_igmp.out_packets, 1);
	NET_COUNT_MIB(icmp_stats.icmpOutMsgs, 1);
	NET_COUNT_MIB(icmp_stats.icmpOutDestUnreachs, 1);
	ip_output(output, TMO_IGMP_OUTPUT);

	return true;
}

/*
*  igmp_timer -- IGMP ʃ^C}[
*/
static void igmp_timer(void *ignore)
{
	T_IGMP_ENTRY *entry = NULL;
	int i, j = 0;

	/* memset(igmp_send_addrs, 0, sizeof(igmp_send_addrs)); */

	syscall(wai_sem(SEM_IGMP_GROUP_LOCK));

	for (i = 0; i < NUM_IGMP_ENTRY; i++) {
		entry = &igmp_groups[i];
		if (entry->timer > 0)
			entry->timer--;
		if (entry->timer == 0) {
			entry->timer = igmp_get_timer(entry);
			igmp_send_addrs[j++] = entry->ip_addr;
		}
	}

	syscall(sig_sem(SEM_IGMP_GROUP_LOCK));

	for (i = 0; i < j; i++) {
		igmp_send_report_v2(igmp_send_addrs[i]);
	}

	timeout(igmp_timer, NULL, NET_TIMER_HZ / 1);
}

/*
*  igmp_init -- IGMP @\B
*/

void igmp_init(void)
{
	T_IGMP_ENTRY *entry = NULL;
	int i;

	for (i = 0; i < NUM_IGMP_ENTRY; i++) {
		igmp_groups[i].timer = -1;
	}

	timeout(igmp_timer, NULL, NET_TIMER_HZ / 1);
}

bool_t igmp_is_joined_group(T_IN4_ADDR addr)
{
	T_IGMP_ENTRY *entry = NULL;
	int		i;
	bool_t ret = false;

	syscall(wai_sem(SEM_IGMP_GROUP_LOCK));

	for (i = 0; i < NUM_IGMP_ENTRY; i++) {
		entry = &igmp_groups[i];
		if ((entry->timer >= 0) && (entry->ip_addr == addr)) {
			ret = true;
			break;
		}
	}

	syscall(sig_sem(SEM_IGMP_GROUP_LOCK));

	return ret;
}

uint16_t igmp_get_timer(T_IGMP_ENTRY *entry)
{
	uint16_t time = entry->resptime;

	if (time == 0)
		time = 10;

	return net_rand() / (0x7fffffff / time);
}

#endif /* of #ifdef SUPPORT_IGMP */
